/*************************************************
		PixelPlus.cpp
		Ver. 0.95

	Copyright (C) 2010-2011,
		MicroFan, Div. PeopleMedia, Inc.
	All rights reserved.
*************************************************/


#include "PixelPlus.h"

#include <stdio.h>
#include <string.h>
#include "WProgram.h"
#include <avr/pgmspace.h>


/*************************************************
* SpiGpio
*************************************************/

SpiGpio::SpiGpio(uint8_t ss, uint8_t addr)
{
    pinSS = ss ;
    digitalWrite(pinSS, HIGH) ;
    pinMode(pinSS, OUTPUT) ; // ~SS

    pinMode(13, OUTPUT) ; // SCK
    pinMode(12, INPUT) ; // MISO
    pinMode(11, OUTPUT) ; // MOSI

    devAddr = addr ;

#if defined(__AVR_ATmega1280__)
    digitalWrite(13, HIGH) ;
    digitalWrite(11, HIGH) ;
#else
    pinMode(10, OUTPUT) ; // SS
/*
    if (!(SPCR & (1 << SPE))) { // ɐݒ肳ĂȂ
        SPCR = (1 << SPE) | (1 << MSTR) ;
        SPSR |= (1 << SPI2X) ;
    }
*/
#endif
}

uint8_t
SpiGpio::access_addr8(uint8_t cmd, uint8_t reg, uint8_t data)
{
#if defined(__AVR_ATmega1280__)
    uint8_t d, so ;
    int i ;

    digitalWrite(pinSS, LOW) ;

    d = cmd | (devAddr << 1) ;
    for (i = 0; i < 8; i++) {
        so = (d & 0x80) ? 0x20:0 ;
        PORTB = (PORTB & 0x5F) | so ; // SCK & MO
        d <<= 1 ;
        PORTB |= 0x80 ; // SCK
    }

    d = reg ;
    for (i = 0; i < 8; i++) {
        so = (d & 0x80) ? 0x20:0 ;
        PORTB = (PORTB & 0x5F) | so ; // SCK & MO
        d <<= 1 ;
        PORTB |= 0x80 ;
    }

    d = data ;
    for (i = 0; i < 8; i++) {
        so = (d & 0x80) ? 0x20:0 ;
        PORTB = (PORTB & 0x5F) | so ; // SCK & MO
        d <<= 1 ;
        PORTB |= 0x80 ;
        d |= (PINB & 0x40) ? 1 : 0 ;
    }

    digitalWrite(pinSS, HIGH) ;

    return d ;
#else
    uint8_t xspcr, xspsr ;

    xspcr = SPCR ;
    xspsr = SPSR ;  
    SPCR = (1 << SPE) | (1 << MSTR) ;
    SPSR |= (1 << SPI2X) ;

    digitalWrite(pinSS, LOW) ;

    SPDR = cmd | (devAddr << 1) ;
    while (!(SPSR & (1 << SPIF)))
        ;
    SPDR = reg ;
    while (!(SPSR & (1 << SPIF)))
        ;
    SPDR = data ;
    while (!(SPSR & (1 << SPIF)))
        ;

    digitalWrite(pinSS, HIGH) ;

    SPCR = xspcr ;
    SPSR = xspsr ;  

    return SPDR ;
#endif
}

inline void
    SpiGpio::write(uint8_t reg, uint8_t data)
{
    access_addr8(GPIO_WRITE, reg, data) ;
}

inline uint8_t
    SpiGpio::read(uint8_t reg)
{
    return access_addr8(GPIO_READ, reg, 0x00) ;
}


/*************************************************
* PixelPlus
*************************************************/

// #include <avr/pgmspace.h>

const uint8_t FONT[] PROGMEM = {

    /*  !"#$%&'()*+,-./ */

    0b00000000,
    0b00000000,
    0b00000000,
    0b00000000,
    0b00000000,

    0b00000000,
    0b00000000,
    0b01001111,
    0b00000000,
    0b00000000,

    0b00000000,
    0b00000111,
    0b00000000,
    0b00000111,
    0b00000000,

    0b00010100,
    0b01111111,
    0b00010100,
    0b01111111,
    0b00010100,

    0b00100100,
    0b00101010,
    0b01111111,
    0b00101010,
    0b00010010,

    0b00100011,
    0b00010011,
    0b00001000,
    0b01100100,
    0b01100010,

    0b00110110,
    0b01001001,
    0b01010101,
    0b00100010,
    0b01010000,

    0b00000000,
    0b00000101,
    0b00000011,
    0b00000000,
    0b00000000,

    0b00000000,
    0b00011100,
    0b00100010,
    0b01000001,
    0b00000000,

    0b00000000,
    0b01000001,
    0b00100010,
    0b00011100,
    0b00000000,

    0b00010100,
    0b00001000,
    0b00111110,
    0b00001000,
    0b00010100,

    0b00001000,
    0b00001000,
    0b00111110,
    0b00001000,
    0b00001000,

    0b00000000,
    0b01010000,
    0b00110000,
    0b00000000,
    0b00000000,

    0b00001000,
    0b00001000,
    0b00001000,
    0b00001000,
    0b00001000,

    0b00000000,
    0b01100000,
    0b01100000,
    0b00000000,
    0b00000000,

    0b00100000,
    0b00010000,
    0b00001000,
    0b00000100,
    0b00000010,


    /* 0123456789:;<=>? */

    0b00111110,
    0b01010001,
    0b01001001,
    0b01000101,
    0b00111110,

    0b00000000,
    0b01000010,
    0b01111111,
    0b01000000,
    0b00000000,

    0b01000010,
    0b01100001,
    0b01010001,
    0b01001001,
    0b01000110,

    0b00100001,
    0b01000001,
    0b01000101,
    0b01001011,
    0b00110001,

    0b00011000,
    0b00010100,
    0b00010010,
    0b01111111,
    0b00010000,

    0b00100111,
    0b01000101,
    0b01000101,
    0b01000101,
    0b00111001,

    0b00111100,
    0b01001010,
    0b01001001,
    0b01001001,
    0b00110000,

    0b00000001,
    0b01110001,
    0b00001001,
    0b00000101,
    0b00000011,

    0b00110110,
    0b01001001,
    0b01001001,
    0b01001001,
    0b00110110,

    0b00000110,
    0b01001001,
    0b01001001,
    0b00101001,
    0b00011110,

    0b00000000,
    0b00110110,
    0b00110110,
    0b00000000,
    0b00000000,

    0b00000000,
    0b01010110,
    0b00110110,
    0b00000000,
    0b00000000,

    0b00001000,
    0b00010100,
    0b00100010,
    0b01000001,
    0b00000000,

    0b00010100,
    0b00010100,
    0b00010100,
    0b00010100,
    0b00010100,

    0b00000000,
    0b01000001,
    0b00100010,
    0b00010100,
    0b00001000,

    0b00000010,
    0b00000001,
    0b01010001,
    0b00001001,
    0b00000110,


    /* @ABCDEFGHIJKLMNO */

    0b00110010,
    0b01001001,
    0b01111001,
    0b01000001,
    0b00111110,

    0b01111110,
    0b00010001,
    0b00010001,
    0b00010001,
    0b01111110,

    0b01111111,
    0b01001001,
    0b01001001,
    0b01001001,
    0b00110110,

    0b00111110,
    0b01000001,
    0b01000001,
    0b01000001,
    0b00100010,

    0b01111111,
    0b01000001,
    0b01000001,
    0b00100010,
    0b00011100,

    0b01111111,
    0b01001001,
    0b01001001,
    0b01001001,
    0b01000001,

    0b01111111,
    0b00001001,
    0b00001001,
    0b00001001,
    0b00000001,

    0b00111110,
    0b01000001,
    0b01001001,
    0b01001001,
    0b00111010,

    0b01111111,
    0b00001000,
    0b00001000,
    0b00001000,
    0b01111111,

    0b00000000,
    0b01000001,
    0b01111111,
    0b01000001,
    0b00000000,

    0b00100000,
    0b01000000,
    0b01000001,
    0b00111111,
    0b00000001,

    0b01111111,
    0b00001000,
    0b00010100,
    0b00100010,
    0b01000001,

    0b01111111,
    0b01000000,
    0b01000000,
    0b01000000,
    0b01000000,

    0b01111111,
    0b00000010,
    0b00000100,
    0b00000010,
    0b01111111,

    0b01111111,
    0b00000100,
    0b00001000,
    0b00010000,
    0b01111111,

    0b00111110,
    0b01000001,
    0b01000001,
    0b01000001,
    0b00111110,


    /* PQRSTUVWXYZ[\]^_ */

    0b01111111,
    0b00001001,
    0b00001001,
    0b00001001,
    0b00000110,

    0b00111110,
    0b01000001,
    0b01010001,
    0b00100001,
    0b01011110,

    0b01111111,
    0b00001001,
    0b00011001,
    0b00101001,
    0b01000110,

    0b01000110,
    0b01001001,
    0b01001001,
    0b01001001,
    0b00110001,

    0b00000001,
    0b00000001,
    0b01111111,
    0b00000001,
    0b00000001,

    0b00111111,
    0b01000000,
    0b01000000,
    0b01000000,
    0b00111111,

    0b00011111,
    0b00100000,
    0b01000000,
    0b00100000,
    0b00011111,

    0b00111111,
    0b01000000,
    0b00111111,
    0b01000000,
    0b00111111,

    0b01100011,
    0b00010100,
    0b00001000,
    0b00010100,
    0b01100011,

    0b00000111,
    0b00001000,
    0b01110000,
    0b00001000,
    0b00000111,

    0b01100001,
    0b01010001,
    0b01001001,
    0b01000101,
    0b01000011,

    0b00000000,
    0b01111111,
    0b01000001,
    0b01000001,
    0b00000000,

    0b00010101,
    0b00010110,
    0b01111100,
    0b00010110,
    0b00010101,

    0b00000000,
    0b01000001,
    0b01000001,
    0b01111111,
    0b00000000,

    0b00000100,
    0b00000010,
    0b00000001,
    0b00000010,
    0b00000100,

    0b01000000,
    0b01000000,
    0b01000000,
    0b01000000,
    0b01000000,


    /* `abcdefghijklmno */

    0b00000000,
    0b00000001,
    0b00000010,
    0b00000100,
    0b00000000,

    0b00100000,
    0b01010100,
    0b01010100,
    0b01010100,
    0b01111000,

    0b01111111,
    0b01001000,
    0b01000100,
    0b01000100,
    0b00111000,

    0b00111000,
    0b01000100,
    0b01000100,
    0b01000100,
    0b00100000,

    0b00111000,
    0b01000100,
    0b01000100,
    0b01001000,
    0b01111111,

    0b00111000,
    0b01010100,
    0b01010100,
    0b01010100,
    0b00011000,

    0b00001000,
    0b01111110,
    0b00001001,
    0b00000010,
    0b00000000,

    0b00001100,
    0b01010010,
    0b01010010,
    0b01010010,
    0b00111110,

    0b01111111,
    0b00001000,
    0b00000100,
    0b00000100,
    0b01111000,

    0b00000000,
    0b01000100,
    0b01111101,
    0b01000000,
    0b00000000,

    0b00100000,
    0b01000000,
    0b01000100,
    0b00111101,
    0b00000000,

    0b01111111,
    0b00010000,
    0b00101000,
    0b01000100,
    0b00000000,

    0b00000000,
    0b01000001,
    0b01111111,
    0b01000000,
    0b00000000,

    0b01111100,
    0b00000100,
    0b01111000,
    0b00000100,
    0b01111000,

    0b01111100,
    0b00001000,
    0b00000100,
    0b00000100,
    0b01111000,

    0b00111000,
    0b01000100,
    0b01000100,
    0b01000100,
    0b00111000,

    /* pqrstuvwxyz<|>__ */

    0b01111100,
    0b00010100,
    0b00010100,
    0b00010100,
    0b00001000,

    0b00011000,
    0b00100100,
    0b00100100,
    0b00011000,
    0b01111100,

    0b01111100,
    0b00001000,
    0b00000100,
    0b00000100,
    0b00001000,

    0b01001000,
    0b01010100,
    0b01010100,
    0b01010100,
    0b00100000,

    0b00000100,
    0b00111111,
    0b01000100,
    0b01000000,
    0b00100000,

    0b00111100,
    0b01000000,
    0b01000000,
    0b00100000,
    0b01111100,

    0b00011100,
    0b00100000,
    0b01000000,
    0b00100000,
    0b00011100,

    0b00111100,
    0b01000000,
    0b00111100,
    0b01000000,
    0b00111100,

    0b01000100,
    0b00101000,
    0b00010000,
    0b00101000,
    0b01000100,

    0b00001100,
    0b01010000,
    0b01010000,
    0b01010000,
    0b00111100,

    0b01000100,
    0b01100100,
    0b01010100,
    0b01001100,
    0b01000100,

    0b00000000,
    0b00001000,
    0b00110110,
    0b01000001,
    0b00000000,

    0b00000000,
    0b00000000,
    0b01111111,
    0b00000000,
    0b00000000,

    0b00000000,
    0b01000001,
    0b00110110,
    0b00001000,
    0b00000000,

    0b00000111,
    0b00000101,
    0b00000111,
    0b00000000,
    0b00000000,

    0b00000010,
    0b00000100,
    0b00001000,
    0b00010000,
    0b00100000,

    /* */

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

    0b01000001,
    0b00110110,
    0b00001000,
    0b00110110,
    0b01000001,

} ;


#define  SPI_GPIO_ADDR  0x00


#define PIXEL_RD  0x01
#define PIXEL_CD  0x02
#define PIXEL_BL  0x04
#define PIXEL_LED  0x08

uint8_t
    PixelPlus::dataGet(uint8_t cd)
{
    uint8_t data ;

    if (ioOut) {
        gpio.write(IODIRA, 0xff) ; // s͂
        ioOut = 0 ;
    }

    vCD = cd ;
//    ctrl = ~PIXEL_RD ;
    ctrl = vCD ? PIXEL_CD:0;
    ctrl |= swBackLight ? PIXEL_BL:0;
    ctrl |= swLed ?  PIXEL_LED:0;
    gpio.write(GPIOB, ctrl) ;
    
    data = gpio.read(GPIOA) ;

    ctrl |= PIXEL_RD ;
    gpio.write(GPIOB, ctrl) ;

    return data ;
}

void
    PixelPlus::dataSet(uint8_t cd, uint8_t data)
{
    uint8_t last ;

    if (!ioOut) {
        gpio.write(IODIRA, 0x00) ; // so͂
        ioOut = 1 ;
    }

    gpio.write(GPIOA, data) ;

    vCD = cd ;
    last = ctrl ;
    ctrl = PIXEL_RD ;
    ctrl |= vCD ? PIXEL_CD:0;
    ctrl |= swBackLight ? PIXEL_BL:0;
    ctrl |= swLed ?  PIXEL_LED:0;
    if (last != ctrl) {
	gpio.write(GPIOB, ctrl) ;
    }

    digitalWrite(pinWR, 0);
    digitalWrite(pinWR, 1);
}

void
    PixelPlus::dataPut(void)
{
    ctrl = PIXEL_RD ;
    ctrl |= vCD ? PIXEL_CD:0;
    ctrl |= swBackLight ? PIXEL_BL:0;
    ctrl |= swLed ?  PIXEL_LED:0;
    gpio.write(GPIOB, ctrl) ;
}

inline uint8_t
    PixelPlus::dataRead(void)
{
    return dataGet(1) ;
}

inline void
    PixelPlus::dataWrite(uint8_t data)
{
    dataSet(1, data) ;
}

inline uint8_t
    PixelPlus::statusRead(void)
{
    return dataGet(0) ;
}

inline void
    PixelPlus::cmdWrite(uint8_t data)
{
    dataSet(0, data) ;
}

PixelPlus::PixelPlus() : gpio(10, SPI_GPIO_ADDR),
    pinWR(8)
{
    pinMode(pinWR, OUTPUT);
    digitalWrite(pinWR, HIGH);
    
    gpio.write(IODIRA, 0xFF) ;	// f[^
    ioOut = 0 ;
    
    gpio.write(IODIRB, 0xF0) ;	// AXCb`
    gpio.write(GPPUB, 0xF0) ;

    vCD = 0 ;
    swBackLight = 0 ;
    swLed = 1 ;
    dataPut() ;
    
    textColor = 0x0000 ; // 
    _bgColor = 0xFFFF ;
    __scrollTop = 0 ;
    __scrollBottom = 15 ;
    swScroll = 1 ;
}


void
    PixelPlus::init()
{
  cmdWrite(0x01); // RESET
//  delay(50);
  delay(120);

  cmdWrite(0x11); //SLEEP OUT
  delay(200);

cmdWrite(0xFF); dataWrite(0x40); dataWrite(0x03); dataWrite(0x1A);
cmdWrite(0xd9); dataWrite(0x60);
cmdWrite(0xc7); dataWrite(0x90);
delay(200);
cmdWrite(0xB1); dataWrite(0x04); dataWrite(0x25); dataWrite(0x18);
cmdWrite(0xB2); dataWrite(0x04); dataWrite(0x25); dataWrite(0x18);
cmdWrite(0xB3); dataWrite(0x04); dataWrite(0x25); dataWrite(0x18); dataWrite(0x04); dataWrite(0x25); dataWrite(0x18);
cmdWrite(0xB4); dataWrite(0x03);
cmdWrite(0xB6); dataWrite(0x15); dataWrite(0x02);

cmdWrite(0xC0); // POWER CONTROL 1 GVDD&VCI1
dataWrite(0x02); dataWrite(0x70);
cmdWrite(0xC1); // POWER CONTROL 2 GVDD&VCI1
dataWrite(0x07);
cmdWrite(0xC2); // POWER CONTROL 3 GVDD&VCI1
dataWrite(0x01); dataWrite(0x01);
cmdWrite(0xC3); // POWER CONTROL 4 GVDD&VCI1
dataWrite(0x02); dataWrite(0x07);
cmdWrite(0xC4); // POWER CONTROL 5 GVDD&VCI1
dataWrite(0x02); dataWrite(0x04);
cmdWrite(0xFC); // POWER CONTROL 6 GVDD&VCI1
dataWrite(0x11); dataWrite(0x17);

cmdWrite(0xC5); //VCOMH&VCOML
dataWrite(0x3c); dataWrite(0x4f);

cmdWrite(0x36); //MV,MX,MY,RGB
dataWrite(0xC8);
//dataWrite(0xA8);
cmdWrite(0x3a); // COLOR MODE
dataWrite(0x05); // 16bit color
//***********************GAMMA*************************
cmdWrite(0xE0); dataWrite(0x06); dataWrite(0x0E); dataWrite(0x05); dataWrite(0x20); dataWrite(0x27); dataWrite(0x23); dataWrite(0x1C); dataWrite(0x21); dataWrite(0x20); dataWrite(0x1C); dataWrite(0x26); dataWrite(0x2F); dataWrite(0x00); dataWrite(0x03); dataWrite(0x00); dataWrite(0x24);
cmdWrite(0xE1); dataWrite(0x06); dataWrite(0x10); dataWrite(0x05); dataWrite(0x21); dataWrite(0x27); dataWrite(0x22); dataWrite(0x1C); dataWrite(0x21); dataWrite(0x1F); dataWrite(0x1D); dataWrite(0x27); dataWrite(0x2F); dataWrite(0x05); dataWrite(0x03); dataWrite(0x00); dataWrite(0x3F);

  cmdWrite(0x29); // DISPLAY ON
//delay(100);
}

void
    PixelPlus::clear()
{
  clear(_bgColor) ;
}

void
    PixelPlus::clear(uint8_t r, uint8_t g, uint8_t b)
{
  uint16_t color ;
  
  color = (r & 0xf8) << 8 ;
  color |= ((g & 0xfc) << 5) ;
  color |= (b >> 3) ;
  
  clear(color) ;
}

void
    PixelPlus::clear(uint16_t color)
{
  uint8_t h, l ;
  
  _bgColor = color ;

  h = color >> 8 ;
  l = color & 0xff ;
  
  cmdWrite(0x2A);
  dataWrite(0x00); dataWrite(0x02);
  dataWrite(0x00); dataWrite(0x81);
  cmdWrite(0x2B);
  dataWrite(0x00); dataWrite(0x03);
  dataWrite(0x00); dataWrite(0x82);

  cmdWrite(0x2C);

  for (int i=0;i<128;i++) {
    for (int j=0;j<128;j++) {
      dataWrite(h);
      if (h == l) {
          digitalWrite(pinWR, 0);
          digitalWrite(pinWR, 1);
      } else
         dataWrite(l);
    }
  }
  
  _col = 0 ;
  _line = 0 ;
}


/***********************************************************
		wiF̐ݒ
***********************************************************/

void
    PixelPlus::setBgColor(uint8_t r, uint8_t g, uint8_t b)
{
  uint16_t color ;
  
  color = (r & 0xf8) << 8 ;
  color |= ((g & 0xfc) << 3) ;
  color |= (b >> 3) ;
  
  setBgColor(color) ;
}

void
    PixelPlus::setBgColor(uint16_t color)
{
  _bgColor = color ;
}


/***********************************************************
		_̕`
***********************************************************/

void
    PixelPlus::pset(uint8_t x, uint8_t y, uint8_t r, uint8_t g, uint8_t b)
{
  uint16_t color ;
  
  color = (r & 0xf8) << 8 ;
  color |= ((g & 0xfc) << 3) ;
  color |= (b >> 3) ;
  
  pset(x, y, color) ;
}

void
    PixelPlus::pset(uint8_t x, uint8_t y, uint16_t pixel)
{
  x += 2 ;
  cmdWrite(0x2a) ;
  dataWrite(0) ; dataWrite(x) ;
  dataWrite(0) ; dataWrite(x) ;
  
  y += 3 ;
  cmdWrite(0x2b) ;
  dataWrite(0) ; dataWrite(y) ;
  dataWrite(0) ; dataWrite(y) ;

  cmdWrite(0x2c) ;
  dataWrite(pixel >> 8) ;
  dataWrite(pixel) ;
}


/***********************************************************
		̕`
***********************************************************/

#define BIT(x)	(1 << (x))

void
    PixelPlus::drawLine(uint8_t x, uint8_t y, uint8_t x1, uint8_t y1, uint8_t r, uint8_t g, uint8_t b)
{
  uint16_t color ;
  
  color = (r & 0xf8) << 8 ;
  color |= ((g & 0xfc) << 3) ;
  color |= (b >> 3) ;
  
  drawLine(x, y, x1, y1, color) ;
}

void
    PixelPlus::drawLine(uint8_t x, uint8_t y, uint8_t x1, uint8_t y1, uint16_t pixel)
{
    int dx, dy, adx, ady, dd, da ;
    uint8_t tmp ;

    if (x > PIXEL_X_MAX)
        return ;

    if (y > PIXEL_Y_MAX)
        return ;

    if (x1 > PIXEL_X_MAX)
        return ;

    if (y1 > PIXEL_Y_MAX)
        return ;

    dx = x - x1 ;
    dy = y - y1 ;

    if (dx == 0) {
        // c̒
        if (y > y1) {
            tmp = y1 ;
            y1 = y ;
            y = tmp ;
        }

        for ( ; y <= y1; y++) {
                pset(x, y, pixel) ;
        }	
    } else if (dy == 0) {
        // ̒
        if (x > x1) {
            tmp = x1 ;
            x1 = x ;
            x = tmp ;
        }

        for ( ; x <= x1; x++) {
                pset(x, y, pixel) ;
        }	
    } else {
        // ΂߂̒
        adx = dx > 0 ? dx : -dx ;
        ady = dy > 0 ? dy : -dy ;
        if (adx > ady) {
            dd = dy << 8 ;
            dd /= dx ;

            if (x > x1) {
                tmp = x1 ;
                x1 = x ;
                x = tmp ;
//                dd = -dd ;
                
                y = y1 ;//?
            }

            da = y << 8 ;
            for ( ; x <= x1; x++) {
                pset(x, da >> 8, pixel) ;
                da += dd ;
            }	
        } else {
            dd = dx << 8 ;
            dd /= dy ;

            if (y > y1) {
                tmp = y1 ;
                y1 = y ;
                y = tmp ;
//                dd = -dd ;
                
                x = x1 ;//?
            }

            da = x << 8 ;
            for ( ; y <= y1; y++) {
                pset(da >> 8, y, pixel) ;
                da += dd ;
            }	
        }	
    }	
}


/***********************************************************
		`̕`
***********************************************************/

void
    PixelPlus::drawRectangle(uint8_t x, uint8_t y, uint8_t x1, uint8_t y1, uint8_t r, uint8_t g, uint8_t b)
{
  uint16_t color ;
  
  color = (r & 0xf8) << 8 ;
  color |= ((g & 0xfc) << 3) ;
  color |= (b >> 3) ;
  
  drawRectangle(x, y, x1, y1, color) ;
}

void
    PixelPlus::drawRectangle(uint8_t x, uint8_t y, uint8_t x1, uint8_t y1, uint16_t pixel)
{
    if (x > PIXEL_X_MAX)
        return ;

    if (y > PIXEL_Y_MAX)
        return ;

    if (x1 > PIXEL_X_MAX)
        return ;

    if (y1 > PIXEL_Y_MAX)
        return ;

    drawLine(x, y, x, y1, pixel) ;
    drawLine(x1, y, x1, y1, pixel) ;
    drawLine(x, y, x1, y, pixel) ;
    drawLine(x, y1, x1, y1, pixel) ;
}	


/***********************************************************
		o͊֌W
***********************************************************/


void
    PixelPlus::setCursor(uint8_t col, uint8_t line)
{
    _col = col ;
    _line = line ;
}

void
    PixelPlus::setTextColor(uint8_t r, uint8_t g, uint8_t b)
{
  uint16_t color ;
  
  color = (r & 0xf8) << 8 ;
  color |= ((g & 0xfc) << 3) ;
  color |= (b >> 3) ;
  
    textColor = color ;
}

void
    PixelPlus::setTextColor(uint16_t color)
{
    textColor = color ;
}


/***********************************************************
		\̃XN[
***********************************************************/


void
    PixelPlus::setScroll(uint8_t top, uint8_t bottom)
{
    if (top >= bottom)
        return ;

    if ((top < 0) || (bottom < 0))
        return ;

    if ((top > 15) || (bottom > 15))
        return ;

    __scrollTop = top ;
    __scrollBottom = bottom ;

    _line = top ;
}


void
    PixelPlus::scroll(uint8_t top, uint8_t bottom)
{
    uint8_t buff[16*3] ;
    uint8_t h, l ;

    cmdWrite(0x3a); // COLOR MODE
    dataWrite(0x06); // 18bit color

    for (int y = top * 8; y < bottom * 8; y++) {
	for (int x = 0; x < 128; x += 16) {
	    int xx, yy ;

	    xx = x + 2 ;
	    yy = y + (8 + 3) ;
	    cmdWrite(0x2a) ;
	    dataWrite(0) ; dataWrite(xx) ;
	    dataWrite(0) ; dataWrite(xx + 15) ;
  
	    cmdWrite(0x2b) ;
	    dataWrite(0) ; dataWrite(yy) ;
	    dataWrite(0) ; dataWrite(yy) ;

	    cmdWrite(0x2e) ;
	    dataRead() ;
	    for (int i = 0; i < 16*3; ) {
		buff[i++] = dataRead() ;
		buff[i++] = dataRead() ;
		buff[i++] = dataRead() ;
	    }

/*
	    xx = x + 2 ;
	    cmdWrite(0x2a) ;
	    dataWrite(0) ; dataWrite(xx) ;
	    dataWrite(0) ; dataWrite(xx + 15) ;
*/
	    yy = y + 3 ;
	    cmdWrite(0x2b) ;
	    dataWrite(0) ; dataWrite(yy) ;
	    dataWrite(0) ; dataWrite(yy) ;

	    cmdWrite(0x2c) ;
	    for (int i = 0; i < 16*3; ) {
		dataWrite(buff[i++]) ;
		dataWrite(buff[i++]) ;
		dataWrite(buff[i++]) ;
	    }
	}
    }

    cmdWrite(0x3a); // COLOR MODE
    dataWrite(0x05); // 16bit color

  h = _bgColor >> 8 ;
  l = _bgColor & 0xff ;
  
  cmdWrite(0x2A);
  dataWrite(0x00); dataWrite(0x02);
  dataWrite(0x00); dataWrite(0x81);
  cmdWrite(0x2B);
  dataWrite(0x00); dataWrite(bottom * 8 +3);
  dataWrite(0x00); dataWrite(bottom * 8 -7+3);

  cmdWrite(0x2C);

  for (int i=0;i<128;i++) {
    for (int j=0;j<8;j++) {
      dataWrite(h);
      if (h == l) {
          digitalWrite(pinWR, 0);
          digitalWrite(pinWR, 1);
      } else
         dataWrite(l);
    }
  }
}


void
    PixelPlus::write(uint8_t c)
{
    uint8_t i, j ;
    uint8_t x, y, d ;

    if (c == '\r') {	// A
        _col = 0 ;
        return ;
    } else if (c == '\n') {	// s
        if (++_line > __scrollBottom) {
          if (swScroll) {
            _line = __scrollBottom ;
            scroll(__scrollTop, __scrollBottom) ;
          } else
            _line = __scrollTop ;
        }	
        return ;
    } else if (c < 32)
        return ;

  x = _col++ * 6 ;
  y = _line * 8 ;

    c -= 32 ;
    for (i = 0; i < 5; i++) {
        //    d = pgm_read_uint8_t(&FONT[c*5 + i]) ;
        d = pgm_read_byte_near(&FONT[c*5 + i]) ;
        for (j = 0; j < 7; j++) {
           if (BIT(j) & d) {
             pset(x + i, y + j, textColor) ;
            }
        }
    }

    if (_col >= 21) {
        _col = 0 ;
        if (++_line > __scrollBottom) {
          if (swScroll) {
            _line = __scrollBottom ;
            scroll(__scrollTop, __scrollBottom) ;
          } else
            _line = __scrollTop ;
        }	
    }
}

/***********************************************************
		obNCg
***********************************************************/


void
    PixelPlus::backLight(uint8_t sw)
{
    swBackLight = sw;
    dataPut() ;
}


/***********************************************************
		uU[
***********************************************************/

void
PixelPlus::beep(uint8_t value)
{
  if (value) {
    analogWrite(9, 100) ;
  } else {
    analogWrite(9, 0) ;
  }
}

/***********************************************************
		LED
***********************************************************/

void
PixelPlus::led(uint8_t value)
{
  if (value) {
    ctrl |= 0x08 ;
  } else {
    ctrl &= ~0x08 ;
  }
  
  gpio.write(GPIOB, ctrl) ;
}


/***********************************************************
		^NgXCb`
***********************************************************/

uint8_t
PixelPlus::sw1(void)
{
  uint8_t v = gpio.read(GPIOB) ;

  return (v & 0x10) ? LOW : HIGH;
}

uint8_t
PixelPlus::sw2(void)
{
  uint8_t v = gpio.read(GPIOB) ;

  return (v & 0x20) ? LOW : HIGH;
}

uint8_t
PixelPlus::sw3(void)
{
  uint8_t v = gpio.read(GPIOB) ;

  return (v & 0x40) ? LOW : HIGH;
}

uint8_t
PixelPlus::sw4(void)
{
  uint8_t v = gpio.read(GPIOB) ;

  return (v & 0x80) ? LOW : HIGH;
}

uint8_t
PixelPlus::swAll(void)
{
  uint8_t v = gpio.read(GPIOB) ;

  return ~v >> 4;
}
