/*************************************************
		PixelPlus.h
		Ver. 0.95

	Copyright (C) 2010-2011,
		MicroFan, Div. PeopleMedia, Inc.
	All rights reserved.
*************************************************/


#ifndef PixelPlus_h
#define PixelPlus_h

#include <inttypes.h>
#include "Print.h"

#define  IODIRA  0x00
#define  IODIRB  0x01
#define  IOCON  0x0A
#define  GPPUA  0x0C
#define  GPPUB  0x0D
#define  GPIOA  0x12
#define  GPIOB  0x13
#define  OLATA  0x14
#define  OLATB  0x15

#define GPIO_READ   0x41
#define GPIO_WRITE  0x40

class SpiGpio {
public:
    SpiGpio(uint8_t, uint8_t) ;
    void write(uint8_t, uint8_t) ;
    uint8_t read(uint8_t) ;
private:
    uint8_t pinSS ;
    uint8_t devAddr ;
    uint8_t access_addr8(uint8_t, uint8_t, uint8_t) ;
} ;


#define	PIXEL_X_MAX	128
#define	PIXEL_Y_MAX	128


#define	PIXEL_PIXEL_CLEAR	0
#define	PIXEL_PIXEL_SET	1
#define	PIXEL_PIXEL_XOR	2


class PixelPlus : public Print {
public:
    PixelPlus() ;

    void init(void) ;
    
    void setBgColor(uint8_t r, uint8_t g, uint8_t b) ;
    void setBgColor(uint16_t color) ;
    void clear(uint8_t r, uint8_t g, uint8_t b) ;
    void clear(uint16_t color) ;
    void clear(void) ;
    void pset(uint8_t, uint8_t, uint8_t r, uint8_t g, uint8_t b) ;
    void pset(uint8_t, uint8_t, uint16_t) ;
    void drawLine(uint8_t, uint8_t, uint8_t, uint8_t, uint8_t r, uint8_t g, uint8_t b) ;
    void drawLine(uint8_t, uint8_t, uint8_t, uint8_t, uint16_t) ;
    void drawRectangle(uint8_t, uint8_t, uint8_t, uint8_t, uint8_t r, uint8_t g, uint8_t b) ;
    void drawRectangle(uint8_t, uint8_t, uint8_t, uint8_t, uint16_t) ;

    void setCursor(uint8_t, uint8_t); 
    void setTextColor(uint8_t r, uint8_t g, uint8_t b) ;
    void setTextColor(uint16_t color) ;
    void setScroll(uint8_t top, uint8_t bottom) ;
    virtual void write(uint8_t);

    void beep(uint8_t);
    void backLight(uint8_t sw) ;
    void led(uint8_t);
    uint8_t sw1(void);
    uint8_t sw2(void);
    uint8_t sw3(void);
    uint8_t sw4(void);
    uint8_t swAll(void);

//    void home();

private:
    SpiGpio gpio ;
    uint8_t pinWR ;
    uint8_t _line, _col ;
    uint8_t swScroll ;
    uint8_t __scrollTop ;
    uint8_t __scrollBottom ;
    uint16_t textColor ;
    uint16_t _bgColor ;
    
    uint8_t vCD ;
    uint8_t swBackLight ;
    uint8_t swLed ;
    uint8_t ioOut ;
    uint8_t ctrl ;

    void scroll(uint8_t top, uint8_t bottom) ;

    uint8_t dataGet(uint8_t cd) ;
    void dataSet(uint8_t cd, uint8_t data) ;
    void dataPut(void) ;

    uint8_t dataRead(void) ;
    void dataWrite(uint8_t data) ;
    uint8_t statusRead(void) ;
    void cmdWrite(uint8_t data) ;
};

#endif
